@@ -1,13 +1,20 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 from decimal import Decimal
 
+from sql import Null
+
+from trytond.i18n import gettext
 from trytond.model import ModelView, ModelSQL, Workflow, fields
 from trytond.wizard import Wizard, StateTransition, StateView, Button
 from trytond.pyson import Eval, Bool, If, Id
 from trytond.pool import Pool
 from trytond.transaction import Transaction
 
+from trytond.modules.product import price_digits
+
+from .exceptions import CostError
+
 __all__ = ['Production', 'AssignFailed', 'Assign']
 
 BOM_CHANGES = ['bom', 'product', 'quantity', 'uom', 'warehouse', 'location',
@@ -17,9 +24,9 @@ BOM_CHANGES = ['bom', 'product', 'quantity', 'uom', 'warehouse', 'location',
 class Production(Workflow, ModelSQL, ModelView):
     "Production"
     __name__ = 'production'
-    _rec_name = 'code'
+    _rec_name = 'number'
 
-    code = fields.Char('Code', select=True, readonly=True)
+    number = fields.Char('Number', select=True, readonly=True)
     reference = fields.Char('Reference', select=1,
         states={
             'readonly': ~Eval('state').in_(['request', 'draft']),
@@ -35,6 +42,17 @@ class Production(Workflow, ModelSQL, ModelView):
             'readonly': Eval('state').in_(['cancel', 'done']),
             },
         depends=['state'])
+    planned_start_date = fields.Date('Planned Start Date',
+        states={
+            'readonly': ~Eval('state').in_(['request', 'draft']),
+            'required': Bool(Eval('planned_date')),
+            },
+        depends=['state', 'planned_date'])
+    effective_start_date = fields.Date('Effective Start Date',
+        states={
+            'readonly': Eval('state').in_(['cancel', 'running', 'done']),
+            },
+        depends=['state'])
     company = fields.Many2One('company.company', 'Company', required=True,
         states={
             'readonly': ~Eval('state').in_(['request', 'draft']),
@@ -46,7 +64,7 @@ class Production(Workflow, ModelSQL, ModelView):
             ],
         states={
             'readonly': (~Eval('state').in_(['request', 'draft'])
-                | Eval('inputs', True) | Eval('outputs', True)),
+                | Eval('inputs', [-1]) | Eval('outputs', [-1])),
             },
         depends=['state'])
     location = fields.Many2One('stock.location', 'Location', required=True,
@@ -55,12 +73,12 @@ class Production(Workflow, ModelSQL, ModelView):
             ],
         states={
             'readonly': (~Eval('state').in_(['request', 'draft'])
-                | Eval('inputs', True) | Eval('outputs', True)),
+                | Eval('inputs', [-1]) | Eval('outputs', [-1])),
             },
         depends=['state'])
     product = fields.Many2One('product.product', 'Product',
         domain=[
-            ('type', '!=', 'service'),
+            ('producible', '=', True),
             ],
         states={
             'readonly': ~Eval('state').in_(['request', 'draft']),
@@ -70,7 +88,8 @@ class Production(Workflow, ModelSQL, ModelView):
             ('output_products', '=', Eval('product', 0)),
             ],
         states={
-            'readonly': ~Eval('state').in_(['request', 'draft']),
+            'readonly': (~Eval('state').in_(['request', 'draft'])
+                | ~Eval('warehouse', 0) | ~Eval('location', 0)),
             'invisible': ~Eval('product'),
             },
         depends=['product'])
@@ -97,28 +116,32 @@ class Production(Workflow, ModelSQL, ModelView):
             'invisible': ~Eval('product'),
             },
         depends=['unit_digits'])
-    cost = fields.Function(fields.Numeric('Cost', digits=(16, 4),
+    cost = fields.Function(fields.Numeric('Cost', digits=price_digits,
             readonly=True), 'get_cost')
     inputs = fields.One2Many('stock.move', 'production_input', 'Inputs',
         domain=[
+            ('shipment', '=', None),
             ('from_location', 'child_of', [Eval('warehouse')], 'parent'),
             ('to_location', '=', Eval('location')),
+            ('company', '=', Eval('company', -1)),
             ],
         states={
             'readonly': (~Eval('state').in_(['request', 'draft', 'waiting'])
-                | ~Eval('location')),
+                | ~Eval('warehouse') | ~Eval('location')),
             },
-        depends=['warehouse', 'location'])
+        depends=['warehouse', 'location', 'company'])
     outputs = fields.One2Many('stock.move', 'production_output', 'Outputs',
         domain=[
+            ('shipment', '=', None),
             ('from_location', '=', Eval('location')),
             ('to_location', 'child_of', [Eval('warehouse')], 'parent'),
+            ('company', '=', Eval('company', -1)),
             ],
         states={
             'readonly': (Eval('state').in_(['done', 'cancel'])
-                | ~Eval('location')),
+                | ~Eval('warehouse') | ~Eval('location')),
             },
-        depends=['warehouse', 'location'])
+        depends=['warehouse', 'location', 'company'])
     state = fields.Selection([
             ('request', 'Request'),
             ('draft', 'Draft'),
@@ -132,11 +155,6 @@ class Production(Workflow, ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(Production, cls).__setup__()
-        cls._error_messages.update({
-                'uneven_costs': ('The costs of the outputs (%(outputs)s) of '
-                    'production "%(production)s" do not match the cost of the '
-                    'production (%(costs)s).')
-                })
         cls._transitions |= set((
                 ('request', 'draft'),
                 ('draft', 'waiting'),
@@ -157,6 +175,7 @@ class Production(Workflow, ModelSQL, ModelView):
                 'cancel': {
                     'invisible': ~Eval('state').in_(['request', 'draft',
                             'assigned']),
+                    'depends': ['state'],
                     },
                 'draft': {
                     'invisible': ~Eval('state').in_(['request', 'waiting',
@@ -164,33 +183,60 @@ class Production(Workflow, ModelSQL, ModelView):
                     'icon': If(Eval('state') == 'cancel',
                         'tryton-clear',
                         If(Eval('state') == 'request',
-                            'tryton-go-next',
-                            'tryton-go-previous')),
+                            'tryton-forward',
+                            'tryton-back')),
+                    'depends': ['state'],
+                    },
+                'reset_bom': {
+                    'invisible': (~Eval('bom')
+                        | ~Eval('state').in_(['request', 'draft', 'waiting'])),
+                    'depends': ['state', 'bom'],
                     },
                 'wait': {
                     'invisible': ~Eval('state').in_(['draft', 'assigned',
                             'waiting', 'running']),
                     'icon': If(Eval('state').in_(['assigned', 'running']),
-                        'tryton-go-previous',
+                        'tryton-back',
                         If(Eval('state') == 'waiting',
                             'tryton-clear',
-                            'tryton-go-next')),
+                            'tryton-forward')),
+                    'depends': ['state'],
                     },
                 'run': {
                     'invisible': Eval('state') != 'assigned',
+                    'depends': ['state'],
                     },
                 'done': {
                     'invisible': Eval('state') != 'running',
+                    'depends': ['state'],
                     },
                 'assign_wizard': {
                     'invisible': Eval('state') != 'waiting',
-                    'readonly': ~Eval('groups', []).contains(
-                        Id('stock', 'group_stock')),
+                    'depends': ['state'],
                     },
                 'assign_try': {},
                 'assign_force': {},
                 })
 
+    @classmethod
+    def __register__(cls, module_name):
+        table_h = cls.__table_handler__(module_name)
+        table = cls.__table__()
+
+        # Migration from 3.8: rename code into number
+        if table_h.column_exist('code'):
+            table_h.column_rename('code', 'number')
+
+        super(Production, cls).__register__(module_name)
+
+        # Migration from 4.0: fill planned_start_date
+        cursor = Transaction().connection.cursor()
+        cursor.execute(*table.update(
+                [table.planned_start_date],
+                [table.planned_date],
+                where=(table.planned_start_date == Null)
+                & (table.planned_date != Null)))
+
     @staticmethod
     def default_state():
         return 'draft'
@@ -214,6 +260,20 @@ class Production(Workflow, ModelSQL, ModelView):
     def default_company():
         return Transaction().context.get('company')
 
+    @fields.depends('planned_date', 'product', 'bom')
+    def on_change_with_planned_start_date(self, pattern=None):
+        if self.planned_date and self.product:
+            if pattern is None:
+                pattern = {}
+            pattern.setdefault('bom', self.bom.id if self.bom else None)
+            for line in self.product.lead_times:
+                if line.match(pattern):
+                    if line.lead_time:
+                        return self.planned_date - line.lead_time
+                    else:
+                        return self.planned_date
+        return self.planned_date
+
     def _move(self, from_location, to_location, company, product, uom,
             quantity):
         Move = Pool().get('stock.move')
@@ -231,47 +291,19 @@ class Production(Workflow, ModelSQL, ModelView):
 
     def _explode_move_values(self, from_location, to_location, company,
             bom_io, quantity):
-        pool = Pool()
-        Move = pool.get('stock.move')
-
         move = self._move(from_location, to_location, company,
             bom_io.product, bom_io.uom, quantity)
         move.from_location = from_location.id if from_location else None
         move.to_location = to_location.id if to_location else None
         move.unit_price_required = move.on_change_with_unit_price_required()
-        values = {}
-        for field_name, field in Move._fields.iteritems():
-            try:
-                value = getattr(move, field_name)
-            except AttributeError:
-                continue
-            if value and field._type in ('many2one', 'one2one'):
-                values[field_name] = value.id
-                values[field_name + '.rec_name'] = value.rec_name
-            else:
-                values[field_name] = value
-        return values
+        return move
 
+    @fields.depends(*BOM_CHANGES)
     def explode_bom(self):
         pool = Pool()
         Uom = pool.get('product.uom')
-        Template = pool.get('product.template')
-        Product = pool.get('product.product')
         if not (self.bom and self.product and self.uom):
-            return {}
-        inputs = {
-            'remove': [r.id for r in self.inputs or []],
-            'add': [],
-            }
-        outputs = {
-            'remove': [r.id for r in self.outputs or []],
-            'add': [],
-            }
-        changes = {
-            'inputs': inputs,
-            'outputs': outputs,
-            'cost': Decimal(0),
-            }
+            return
 
         if self.warehouse:
             storage_location = self.warehouse.storage_location
@@ -280,64 +312,45 @@ class Production(Workflow, ModelSQL, ModelView):
 
         factor = self.bom.compute_factor(self.product, self.quantity or 0,
             self.uom)
+        inputs = []
         for input_ in self.bom.inputs:
             quantity = input_.compute_quantity(factor)
-            values = self._explode_move_values(storage_location, self.location,
+            move = self._explode_move_values(storage_location, self.location,
                 self.company, input_, quantity)
-            if values:
-                inputs['add'].append((-1, values))
+            if move:
+                inputs.append(move)
                 quantity = Uom.compute_qty(input_.uom, quantity,
-                    input_.product.default_uom)
-                changes['cost'] += (Decimal(str(quantity)) *
-                    input_.product.cost_price)
+                    input_.product.default_uom, round=False)
+        self.inputs = inputs
 
-        if hasattr(Product, 'cost_price'):
-            digits = Product.cost_price.digits
-        else:
-            digits = Template.cost_price.digits
+        outputs = []
         for output in self.bom.outputs:
             quantity = output.compute_quantity(factor)
-            values = self._explode_move_values(self.location, storage_location,
+            move = self._explode_move_values(self.location, storage_location,
                 self.company, output, quantity)
-            if values:
-                values['unit_price'] = Decimal(0)
-                if output.product.id == values.get('product') and quantity:
-                    values['unit_price'] = Decimal(
-                        changes['cost'] / Decimal(str(quantity))
-                        ).quantize(Decimal(str(10 ** -digits[1])))
-                outputs['add'].append((-1, values))
-        return changes
+            if move:
+                move.unit_price = Decimal(0)
+                outputs.append(move)
+        self.outputs = outputs
 
     @fields.depends('warehouse')
     def on_change_warehouse(self):
-        changes = {
-            'location': None,
-            }
+        self.location = None
         if self.warehouse:
-            changes['location'] = self.warehouse.production_location.id
-        return changes
+            self.location = self.warehouse.production_location
 
-    @fields.depends(*BOM_CHANGES)
+    @fields.depends('product', 'uom', methods=['explode_bom'])
     def on_change_product(self):
-        result = {}
         if self.product:
-            uoms = self.product.default_uom.category.uoms
-            if (not self.uom or self.uom not in uoms):
-                result['uom'] = self.product.default_uom.id
-                result['uom.rec_name'] = self.product.default_uom.rec_name
-                result['unit_digits'] = self.product.default_uom.digits
+            category = self.product.default_uom.category
+            if not self.uom or self.uom.category != category:
+                self.uom = self.product.default_uom
+                self.unit_digits = self.product.default_uom.digits
         else:
-            result['bom'] = None
-            result['uom'] = None
-            result['uom.rec_name'] = ''
-            result['unit_digits'] = 2
-
-        if 'uom' in result:
-            self.uom = result['uom']
-        if 'bom' in result:
-            self.bom = result['bom']
-        result.update(self.explode_bom())
-        return result
+            self.bom = None
+            self.uom = None
+            self.unit_digits = 2
+        self.explode_bom()
 
     @fields.depends('product')
     def on_change_with_uom_category(self, name=None):
@@ -350,17 +363,21 @@ class Production(Workflow, ModelSQL, ModelView):
             return self.uom.digits
         return 2
 
-    @fields.depends(*BOM_CHANGES)
+    @fields.depends(methods=['explode_bom'])
     def on_change_bom(self):
-        return self.explode_bom()
+        self.explode_bom()
 
-    @fields.depends(*BOM_CHANGES)
+    @fields.depends(methods=['explode_bom'])
     def on_change_uom(self):
-        return self.explode_bom()
+        self.explode_bom()
 
-    @fields.depends(*BOM_CHANGES)
+    @fields.depends(methods=['explode_bom'])
     def on_change_quantity(self):
-        return self.explode_bom()
+        self.explode_bom()
+
+    @ModelView.button_change(*BOM_CHANGES)
+    def reset_bom(self):
+        self.explode_bom()
 
     def get_cost(self, name):
         cost = Decimal(0)
@@ -394,10 +411,6 @@ class Production(Workflow, ModelSQL, ModelView):
         return cost
 
     def set_moves(self):
-        pool = Pool()
-        Template = pool.get('product.template')
-        Product = pool.get('product.product')
-
         storage_location = self.warehouse.storage_location
         location = self.location
         company = self.company
@@ -414,7 +427,6 @@ class Production(Workflow, ModelSQL, ModelView):
             return
 
         factor = self.bom.compute_factor(self.product, self.quantity, self.uom)
-        cost = Decimal(0)
         for input_ in self.bom.inputs:
             quantity = input_.compute_quantity(factor)
             product = input_.product
@@ -423,12 +435,6 @@ class Production(Workflow, ModelSQL, ModelView):
             if move:
                 move.production_input = self
                 move.save()
-                cost += Decimal(str(quantity)) * product.cost_price
-
-        if hasattr(Product, 'cost_price'):
-            digits = Product.cost_price.digits
-        else:
-            digits = Template.cost_price.digits
 
         for output in self.bom.outputs:
             quantity = output.compute_quantity(factor)
@@ -437,34 +443,56 @@ class Production(Workflow, ModelSQL, ModelView):
                 product, output.uom, quantity)
             if move:
                 move.production_output = self
-                if product == self.product:
-                    move.unit_price = Decimal(
-                        cost / Decimal(str(quantity))
-                        ).quantize(Decimal(str(10 ** -digits[1])))
-                else:
-                    move.unit_price = Decimal(0)
+                move.unit_price = Decimal(0)
                 move.save()
         self._set_move_planned_date()
 
+    @classmethod
+    def set_cost(cls, productions):
+        pool = Pool()
+        Uom = pool.get('product.uom')
+        Move = pool.get('stock.move')
+
+        digits = Move.unit_price.digits
+        digit = Decimal(str(10 ** -digits[1]))
+        moves = []
+        for production in productions:
+            if not production.quantity or not production.uom:
+                continue
+            if production.company.currency.is_zero(
+                    production.cost - production.output_cost):
+                continue
+            unit_price = production.cost / Decimal(str(production.quantity))
+            for output in production.outputs:
+                if output.product == production.product:
+                    output.unit_price = Uom.compute_price(
+                        production.uom, unit_price, output.uom).quantize(digit)
+                    moves.append(output)
+        Move.save(moves)
+
     @classmethod
     def validate(cls, productions):
         super(Production, cls).validate(productions)
         for production in productions:
             production.check_cost()
 
+    @property
+    def output_cost(self):
+        cost_price = Decimal(0)
+        for output in self.outputs:
+            cost_price += (Decimal(str(output.quantity)) * output.unit_price)
+        return cost_price
+
     def check_cost(self):
         if self.state != 'done':
             return
-        cost_price = Decimal(0)
-        for output in self.outputs:
-            cost_price += (Decimal(str(output.quantity))
-                * output.unit_price)
+        cost_price = self.output_cost
         if not self.company.currency.is_zero(self.cost - cost_price):
-            self.raise_user_error('uneven_costs', {
-                    'production': self.rec_name,
-                    'costs': self.cost,
-                    'outputs': cost_price,
-                    })
+            raise CostError(gettext(
+                    'production.msg_uneven_costs',
+                    production=self.rec_name,
+                    costs=self.cost,
+                    outputs=cost_price))
 
     @classmethod
     def create(cls, vlist):
@@ -474,7 +502,9 @@ class Production(Workflow, ModelSQL, ModelView):
         vlist = [x.copy() for x in vlist]
         config = Config(1)
         for values in vlist:
-            values['code'] = Sequence.get_id(config.production_sequence.id)
+            if values.get('number') is None:
+                values['number'] = Sequence.get_id(
+                    config.production_sequence.id)
         productions = super(Production, cls).create(vlist)
         for production in productions:
             production._set_move_planned_date()
@@ -486,9 +516,18 @@ class Production(Workflow, ModelSQL, ModelView):
         for production in sum(args[::2], []):
             production._set_move_planned_date()
 
+    @classmethod
+    def copy(cls, productions, default=None):
+        if default is None:
+            default = {}
+        else:
+            default = default.copy()
+        default.setdefault('number', None)
+        return super(Production, cls).copy(productions, default=default)
+
     def _get_move_planned_date(self):
         "Return the planned dates for input and output moves"
-        return self.planned_date, self.planned_date
+        return self.planned_start_date, self.planned_date
 
     def _set_move_planned_date(self):
         "Set planned date of moves for the shipments"
@@ -543,7 +582,11 @@ class Production(Workflow, ModelSQL, ModelView):
     def run(cls, productions):
         pool = Pool()
         Move = pool.get('stock.move')
+        Date = pool.get('ir.date')
         Move.do([m for p in productions for m in p.inputs])
+        cls.write([p for p in productions if not p.effective_start_date], {
+                'effective_start_date': Date.today(),
+                })
 
     @classmethod
     @ModelView.button
@@ -552,6 +595,7 @@ class Production(Workflow, ModelSQL, ModelView):
         pool = Pool()
         Move = pool.get('stock.move')
         Date = pool.get('ir.date')
+        cls.set_cost(productions)
         Move.do([m for p in productions for m in p.outputs])
         cls.write([p for p in productions if not p.effective_date], {
                 'effective_date': Date.today(),
@@ -559,7 +603,7 @@ class Production(Workflow, ModelSQL, ModelView):
 
     @classmethod
     @ModelView.button_action('production.wizard_assign')
-    def assign_wizard(self, productions):
+    def assign_wizard(cls, productions):
         pass
 
     @classmethod
@@ -607,13 +651,13 @@ class Assign(Wizard):
     start = StateTransition()
     failed = StateView('production.assign.failed',
         'production.assign_failed_view_form', [
-            Button('Force Assign', 'force', 'tryton-go-next',
+            Button('Force Assign', 'force', 'tryton-forward',
                 states={
                     'invisible': ~Id('stock',
                         'group_stock_force_assignment').in_(
                         Eval('context', {}).get('groups', [])),
                     }),
-            Button('Ok', 'end', 'tryton-ok', True),
+            Button('OK', 'end', 'tryton-ok', True),
             ])
     force = StateTransition()
 
